#ifndef __NETWORK_H
#define __NETWORK_H

#include "definitions.h"

#ifdef __NETWORK_SUPPORT__

#include "SDL_net.h"

#include "console.h"

/*
	demo tree
	=========
	
	node server	1 (master only)
		|
		+------ client 1 (client only)
		|
		+------ client 2 (client only)
		|
		+------ node server 2 (client/master)
		|           |
		|           +------ client 1 (client only)
		|           |
		|           +------ client 2 (client only)
		|
		+------ node server 3 (client/master)
		|           |
		|           +------ client 1 (client only)
		|           |
		|           +------ client 2 (client only)
		.
		.
	
	and how does a packet look like?
	"sender id:next node id:next id:next id:target id:MESSAGE:some text"
	it will cover the whole route through the tree from the
	sender to the recepient. then follows the packet-type
	( for ex. MESSAGE or BROADCAST or UNICAST ) and some text if needed.
	
	broadcasting and unicasting (searching for buddies):
	broadcast packets will be sent once in a while for example
	when first connecting to the tree or updating the buddylist.
	it will look like this:
	"sender id:next node id:next node id:...:BROADCAST:sender aliasname"
	the idea is that the sender will create a packet like this
	"sender id:BROADCAST:sender aliasname"
	and send it to his clients and next master/node server.
	this will then append its id like this
	"sender id:server id:BROADCAST:sender aliasname"
	the recepient will do the same until one knows noone
	else and create a packet like this
	"my id:id:id:...:id:id:former sender id:UNICAST:my aliasname"
	so this will cover the hole route from the recepient to
	the former creator of the bc packet.
	everyone who gets this packet looks into this and
	searches for its id in this message. if it finds itself
	in this message, it will think that this packet is on
	the correct route and send it to the following id.
	(but only when the follower exists in its peer list!!!)
	after some time all packets should arrive at the former
	sender and it will be able to put those people into its
	buddylist. note that every client or server will be able
	to put this sender into it's buddylist too because it
	already contains the neseccary information to contact
	and name it! (the last argument covers the aliasname)
	also note that every master/node server will also create
	an unicast message for the former sender.
*/

class Network
{
	public:
	struct peer
	{
		int       type          ;
		char      name  [ 256 ] ;
		char      alias [ 256 ] ;
		int       port          ;
		TCPsocket socket        ;
		IPaddress peer          ;
		bool      established   ;
		Uint32    timeout       ;
		Uint32    born          ;
	} ;
	
	struct packet
	{
		uint                   type     ;
		Uint32                 from_ip  ;
		Uint32                 dest_ip  ;
		Uint32                 hash     ;
		uint                   hops     ;
		uint                   max_hops ;
		GlDynArray < Uint32 >  route    ;
		char                  *data     ;
	} ;
	
	enum
	{
		NM_LOCAL,
		NM_SERVER,
		NM_CLIENT
	} ;
	
	enum
	{
		NM_SYSTEM_MESSAGE,		// network related stuff
		NM_PLAYFIELD_MESSAGE,	// should be send to the playfield
		NM_USER_MESSAGE,		// play chat / nothing from above / nonesense
		NM_PING_MESSAGE
	} ;
	
	void Init () ;
	
	void EnableConsolePrinting ( Console *new_console, Console *new_buddy_list ) ;
	void DisableConsolePrinting () ;

	void ResolveIP ( IPaddress *ip, char *hostname ) ;
	
	void Printf ( const char *format, ... ) ;
	
	std::string AddUsernameAtBeginning ( std::string text ) ;
	
	int SendMessage ( peer *peer, char *message ) ;
	
	int  FindPeerBySocketI ( TCPsocket socket ) ;
	peer *FindPeerBySocket ( TCPsocket socket ) ;
	
	void ConnectTo ( char *client, int port ) ;
	void ListenOnPort ( int port ) ;
	void Disconnect ( peer *peer ) ;

	/* obsolete */
	void StartAsServer ( int port ) ;
	void StartAsClient ( char *server, int port ) ;
	int  Connect () ;
	void Disconnect () ;
	/*          */
	
	void TCPAcceptEvent ( TCPsocket socket ) ;
	bool TCPReceiveEvent ( peer *peer ) ; // returns true = alive ; false = died
	void TCPCloseEvent ( peer *peer ) ;
	
	void UpdateBuddyList () ;
	
	void ProcessEvents ( SDL_Event *event ) ;
	
	char *StripHeadAndTail ( char *text ) ;
	char *StripMarks ( char *text ) ;
	GlDynArray < std::string > *GetArgList ( char *string,
												char delimiter = ' ' ) ;
	int ProcessMessage ( peer *peer, char *t_message ) ;
	
	enum
	{
		NM_ALLDATA,
		NM_ALIASNAME
	} ;
	void SendUserData ( peer *peer = NULL, int type = NM_ALLDATA ) ;
	
	void Quit () ;
	
	private:
	GlDynArray < peer > *peers ;
	Console *console ;
	Console *buddy_list ;
	std::string aliasname ;
	IPaddress serverIP ;
	TCPsocket servsock ;
	SDLNet_SocketSet socketset ;
	GlDynArray < peer > *seen_users ;
} ;

#endif

#endif
