#include "network.h"

#ifdef __NETWORK_SUPPORT__

#include <stdarg.h>
#include <unistd.h>

extern char pkg_filename[] ;

extern bool dedicated_server ;
bool network_debugging = false ;

int MAX_CONNECTIONS = 256 ; // max number of connections

int MAX_HOPS = 30 ; // max number of hosts a message can come thru

char PACKET_DATA_DELIMITER = ';' ;

int BUDDY_TIMEOUT      = 120 ; // seconds 'till a buddy is removed
int CONNECTION_TIMEOUT = 60 ; // seconds 'till a socket is beeing closed

int HOST_CACHE_TIMEOUT = 60 ; // seconds while hosts are remembered

int PING_INTERVAL = 10 ; // interval in seconds for life-tracking

void Network::Init ()
{
	peer_list = new GlDynArray < peer > () ;
	peer_list -> Clear () ;
	
	known_packets = new GlDynArray < known_packet > () ;
	known_packets -> Clear () ;
	
	buddies = new GlDynArray < host > () ;
	buddies -> Clear () ;
	
	console = NULL ;
	buddy_list = NULL ;
	
	SDL_EnableUNICODE ( 1 ) ;
	if ( SDLNet_Init () )
	{
		printd ( "Failed to initialize SDL_Net error=%s\n", SDL_GetError () ) ;
		exit ( 1 ) ;
	}
	
	// Allocate the socket set
	socketset = SDLNet_AllocSocketSet(MAX_CONNECTIONS+1);
	if ( socketset == NULL )
	{
		fprintf(stderr, "Couldn't create socket set: %s\n",
		SDLNet_GetError());
		exit(2);
	}

	//self . ip = NULL ; how do this? perhaps set host and port to 0?
	self . socket = NULL ;
	
	if ( strlen ( getenv ( "USER" ) ) > 0 )
		sprintf ( self . alias, "%s", getenv ( "USER" ) ) ;
	else if ( strlen ( getenv ( "HOST" ) ) > 0 )
		sprintf ( self . alias, "%s", getenv ( "HOST" ) ) ;
	else if ( strlen ( getenv ( "HOSTNAME" ) ) > 0 )
		sprintf ( self . alias, "%s", getenv ( "HOSTNAME" ) ) ;
	else if ( strlen ( rindex ( getenv ( "HOME" ), '/' ) ) > 0 )
		sprintf ( self . alias, "%s", rindex ( getenv ( "HOME" ), '/' ) ) ;
	else
		sprintf ( self . alias, "%s", "you" ) ;
	
	srand ( SDL_GetTicks () + self . ip . port + (int) self . alias +
			rand () - SDL_GetTicks () + rand () % SDL_GetTicks () ) ;
	
	sprintf ( self . key, "%s", CreateKey () . c_str () ) ;
	
	//printd ( "my unique key-id (%d chars) is '%s'\n", 
	//			strlen ( self . key ), self . key ) ;
	
	connectto_list . Clear () ;

	dnscache . Clear () ;
}

void Network::EnableDebugging ()
{
	network_debugging = true ;
}

void Network::DisableDebugging ()
{
	network_debugging = false ;
}

void Network::AddHostToConnectionList ( simple_host *host )
{
	connectto_list . PushBack ( *host ) ;
}

void Network::ConnectToServersInList ()
{
	for ( uint i = 0 ; i < connectto_list . Count () ; i++ )
	{
		simple_host *host = connectto_list . ItemPointer ( i ) ;
		ConnectTo ( host -> address, host -> port ) ;
	}
}

void Network::EnableConsolePrinting ( Console *new_console, Console *new_buddy_list )
{
	console = new_console ;
	buddy_list = new_buddy_list ;
	
	if ( buddy_list )
		UpdateBuddyList () ;
	
#ifdef __NCURSES_SUPPORT__
	if ( dedicated_server )
	{
		nconsole = new NConsole () ;
		
		ProcessMessage ( NULL, "//about" ) ;
		ProcessMessage ( NULL, "//help" ) ;

		ConnectToServersInList () ;
	}
#endif
	if ( console )
	{
		ProcessMessage ( NULL, "//about" ) ;
		ProcessMessage ( NULL, "//help" ) ;

		if ( strcmp ( pkg_filename, "network" ) == 0 )
			ConnectToServersInList () ;
	}
}

void Network::DisableConsolePrinting ()
{
	console = NULL ;
	buddy_list = NULL ;
}

const char *Network::Cached_SDLNet_ResolveIP ( IPaddress *ip )
{
	const char *hostname = NULL ;

	if ( ip == NULL )
		return NULL ;
	
	for ( uint i = 0 ; i < dnscache . Count () ; i++ )
	{
		if ( dnscache [ i ] . ip . host == ip -> host &&
			dnscache [ i ] . ip . port == ip -> port )
		{
			hostname = dnscache [ i ] . alias ;
			//printd ( "got hostname '%s' from dns-cache.\n", hostname ) ;
		}
	}
	
	if ( ! hostname )
	{
		hostname = SDLNet_ResolveIP ( ip ) ;

		peer peer ;
		peer . ip . host = ip -> host ;
		peer . ip . port = ip -> port ;
		peer . timeout = SDL_GetTicks () + PING_INTERVAL * 1000 ;
		sprintf ( peer . alias, "%s", hostname ) ;
		dnscache . PushBack ( peer ) ;

		//printd ( "added hostname '%s' to dns-cache.\n", hostname ) ;
	}

	return hostname ;
}

int Network::Cached_SDLNet_ResolveHost ( IPaddress *address,
						const char *host,
						Uint16 port )
{
	int retval = 0 ;
	
	for ( uint i = 0 ; i < dnscache . Count () ; i++ )
	{
		if ( host != NULL &&
			strcmp ( dnscache [ i ] . alias, host ) == 0 &&
			dnscache [ i ] . ip . port == port )
		{
			address -> host = dnscache [ i ] . ip . host ;
			address -> port = dnscache [ i ] . ip . port ;
			//printd ( "got ip for hostname '%s:%d' from dns-cache.\n", dnscache [ i ] . alias, port ) ;
			return dnscache [ i ] . born ;
		}
	}
	
	retval = SDLNet_ResolveHost ( address, (char*) host, port ) ;

	peer peer ;
	peer . ip . host = address -> host ;
	peer . ip . port = address -> port ;
	sprintf ( peer . alias, "%s", host ) ;
	peer . born = retval ;
	peer . timeout = SDL_GetTicks () + PING_INTERVAL * 1000 ;
	dnscache . PushBack ( peer ) ;

	//printd ( "added ip for hostname '%s:%d' to cache.\n", host, port ) ;

	return retval ;
}

void Network::GetIPAsString ( IPaddress *ip, char *return_string )
{
	// get ip address
	int m1 = 0, m2 = 0, m3 = 0, m4 = 0 ;
	char hips [ 16 ] ;
	
	sprintf ( hips, "%08x", ip -> host ) ;
	m4 = hex_string_to_int ( hips, 0, 2 ) ;
	m3 = hex_string_to_int ( hips, 2, 2 ) ;
	m2 = hex_string_to_int ( hips, 4, 2 ) ;
	m1 = hex_string_to_int ( hips, 6, 2 ) ;
	sprintf ( return_string, "%d.%d.%d.%d", m1, m2, m3, m4 ) ;
}

void Network::ResolveIP ( IPaddress *ip, char *return_hostname )
{
	char *ip_name = (char*) Cached_SDLNet_ResolveIP ( ip ) ;
	if ( ip_name == NULL )
		GetIPAsString ( ip, return_hostname ) ;
	else	// get hostname
		sprintf ( return_hostname, "%s", ip_name ) ;
}

char *Network::ResolveIP ( char *address )
{
	IPaddress ip ;
	if ( Cached_SDLNet_ResolveHost ( &ip, address, 0 ) != -1 )
		return (char*) Cached_SDLNet_ResolveIP ( &ip ) ;
	else
		return address ;
}

void Network::printd ( const char *format, ... )
{	
	if ( network_debugging == false ) return ;
	
	// this one is copied from kyra's console.cpp file :)
	// slightly modified by me to use either a KrConsole or
	// the stdout stream...
	va_list va ;
	char buffer [ 8192 ] ;
	
	//
	//  format and output the message..
	//
	va_start ( va, format ) ;
	vsprintf ( buffer, format, va ) ;
	va_end ( va ) ;
	
	char *start ;
	char *end ;
	char *next ;
	
	start = buffer ;
	
	while ( start && *start )
	{
		end = strchr ( start, '\n' ) ;
		if ( end )
		{
			next  = end + 1 ;
			*end = 0 ;
#ifdef __NCURSES_SUPPORT__
			if ( nconsole )
				nconsole -> Printf ( "| %s\n", start ) ;
			else
#endif
				printf ( "| %s\n", start ) ;
			start = next ;
		}
		else
		{
#ifdef __NCURSES_SUPPORT__
			if ( nconsole )
				nconsole -> Printf ( "| %s\n", start ) ;
			else
#endif
				printf ( "%s", start ) ;
			start = 0 ;
		}
	}
}

void Network::Printf ( const char *format, ... )
{	// this one is copied from kyra's console.cpp file :)
	// slightly modified by me to use either a KrConsole or
	// the stdout stream...
	va_list va ;
	char buffer [ 8192 ] ;
	
	//
	//  format and output the message..
	//
	va_start ( va, format ) ;
	vsprintf ( buffer, format, va ) ;
	va_end ( va ) ;
	
	char *start ;
	char *end ;
	char *next ;
	
	start = buffer ;
	
	while ( start && *start )
	{
		end = strchr ( start, '\n' ) ;
		if ( end )
		{
			next  = end + 1 ;
			*end = 0 ;
			if ( console != NULL )
				console -> PushText ( start ) ;
#ifdef __NCURSES_SUPPORT__
			else if ( nconsole != NULL )
				nconsole -> PushText ( start ) ;
#endif
			else
				printf ( "%s\n", start ) ;
			start = next ;
		}
		else
		{
			if ( console != NULL )
				console -> PushText ( start ) ;
#ifdef __NCURSES_SUPPORT__
			else if ( nconsole != NULL )
				nconsole -> PushText ( start ) ;
#endif
			else
				printf ( "%s", start ) ;
			start = 0 ;
		}
	}
}

std::string Network::AddUsernameAtBeginning ( std::string text )
{
	std::string t_text ;
	
	t_text = self . alias ;
	t_text += ": " ;
	t_text += text ;
	
	return t_text ;
}

bool Network::AddressIsKnownToPeer ( peer *peer, char *address )
{
	GlDynArray < host > *known_hosts = &peer -> known_hosts ;
	
	for ( unsigned int i = 0 ; i < known_hosts -> Count () ; i++ )
	{
		host *host = known_hosts -> ItemPointer ( i ) ;
		
		if ( strcmp ( host -> address, address ) == 0 )
			return true ;
	}
	
	return false ;
}

bool Network::KeyIsKnownToPeer ( peer *peer, char *key )
{
	GlDynArray < host > *known_hosts = &peer -> known_hosts ;
	
	for ( unsigned int i = 0 ; i < known_hosts -> Count () ; i++ )
	{
		host *host = known_hosts -> ItemPointer ( i ) ;
		
		if ( strcmp ( host -> key, key ) == 0 )
			return true ;
	}
	
	return false ;
}

int Network::SendPacket ( peer *peer, packet *packet )
{
	if ( ! peer ) return 0 ;
	
	char data [ 8192 ] ;
	sprintf ( data, "%d%c", packet -> type, PACKET_DATA_DELIMITER ) ;
	sprintf ( data, "%s%s%c", data, packet -> from, PACKET_DATA_DELIMITER ) ;
	sprintf ( data, "%s%s%c", data, packet -> dest, PACKET_DATA_DELIMITER ) ;
	sprintf ( data, "%s%s%c", data, packet -> from_key, PACKET_DATA_DELIMITER ) ;
	sprintf ( data, "%s%s%c", data, packet -> dest_key, PACKET_DATA_DELIMITER ) ;
	sprintf ( data, "%s%s%c", data, packet -> alias, PACKET_DATA_DELIMITER ) ;
	sprintf ( data, "%s%s%c", data, packet -> hash, PACKET_DATA_DELIMITER ) ;
	sprintf ( data, "%s%d%c", data, packet -> hops, PACKET_DATA_DELIMITER ) ;
	sprintf ( data, "%s%d%c", data, packet -> max_hops, PACKET_DATA_DELIMITER ) ;
	sprintf ( data, "%s%s", data, packet -> data ) ;
	for ( int i = strlen ( data ) ; i < 8192 ; i++ )
		data [ i ] = '\0' ;
	data [ 8191 ] = '\0' ;
	
	//printd ( "sending packet:\n%s\n", data ) ;
	
	int num_chars = SDLNet_TCP_Send ( peer -> socket, data,
										min ( strlen ( data ) + 1, 8192 ) ) ;
	
	return num_chars ;
}

int Network::SendMessage ( char *address, char *message, uint packet_type )
{
	struct packet packet ;
	int num_chars = 0 ;
	
	// build packet for delivery
	packet . type = packet_type ;
	sprintf ( packet . hash, "%s", CreateKey () . c_str () ) ;
	packet . hops = 0 ;
	packet . max_hops = MAX_HOPS ;
	sprintf ( packet . from_key, "%s", self . key ) ;
	sprintf ( packet . alias, "%s", self . alias ) ;
	sprintf ( packet . data, "%s", message ) ;
	
	for ( unsigned int i = 0 ; i < peer_list -> Count () ; i++ )
	{
		peer *peer = peer_list -> ItemPointer ( i ) ;
		
		char dest [ 256 ] ;
		
		if ( address == NULL )
			sprintf ( dest, "%s", peer -> address ) ;
		else
			sprintf ( dest, "%s", address ) ;
		
		host host ;
		sprintf ( host . key, "%s", peer -> key ) ;
		if ( AddressIsKnownToPeer ( peer , dest ) )
			for ( unsigned int j = 0 ; j < peer -> known_hosts . Count () ; j++ )
				if ( strcmp ( peer -> known_hosts . Item ( j ) . address, dest ) == 0 )
					sprintf ( host . key, "%s", peer -> known_hosts . Item ( j ) . key ) ;
		
		if ( strcmp ( peer -> address, dest ) == 0 ||
				AddressIsKnownToPeer ( peer, dest ) )
		{
			sprintf ( packet . from, "%s", peer -> outgoing_name ) ;
			sprintf ( packet . dest_key, "%s", host . key ) ;
			sprintf ( packet . dest, "%s", dest ) ;
			num_chars += SendPacket ( peer, &packet ) ;
		}
	}
	
	return num_chars ;
}

int Network::FindPeerBySocketI ( TCPsocket socket )
{
	for ( unsigned int i = 0 ; i < peer_list -> Count () ; i++ )
		if ( peer_list -> Item ( i ) . socket == socket )
			return i ;
	
	return -1 ;
}

Network::peer *Network::FindPeerBySocket ( TCPsocket socket )
{
	for ( unsigned int i = 0 ; i < peer_list -> Count () ; i++ )
		if ( peer_list -> Item ( i ) . socket == socket )
			return peer_list -> ItemPointer ( i ) ;
	
	return NULL ;
}

void Network::ConnectTo ( char *client, int port )
{
	Printf ( "|| connecting to %s at port %d...\n",
				client, port ) ;
	
	IPaddress ip ;
	Cached_SDLNet_ResolveHost(&ip, client, port);
	TCPsocket socket = SDLNet_TCP_Open ( &ip ) ;
	if ( ip.host == INADDR_NONE || socket == NULL )
	{
		Printf ( "|| can't connect to %s at port %d !\n",
					client, port ) ;
	}
	else
	{
		struct peer t_peer ;
		t_peer . socket = socket ;
		Cached_SDLNet_ResolveHost(&t_peer . ip, client, port);
		GetIPAsString ( &t_peer . ip, (char*) t_peer . address ) ;
		t_peer . established = false ;
		t_peer . timeout = SDL_GetTicks () + CONNECTION_TIMEOUT * 1000 ;
		t_peer . born = SDL_GetTicks () ;
		t_peer . known_hosts . Clear () ;
		t_peer . outgoing_name [ 0 ] = '\0' ;
		t_peer . key [ 0 ] = '\0' ;
		t_peer . alias [ 0 ] = '\0' ;
		peer_list -> PushBack ( t_peer ) ;
		
		SDLNet_TCP_AddSocket(socketset, socket);
		Printf ( "|| successfully connected to %s:%d. waiting for response...\n",
					client, port ) ;
	}
}

void Network::ListenOnPort ( int port )
{
	// Create the server socket
	Cached_SDLNet_ResolveHost(&self . ip, NULL, port);
	self . socket = SDLNet_TCP_Open(&self . ip);
	if ( self . socket == NULL )
	{
		//serverIP = NULL ;
		Printf( "|| couldn't create server socket: %s\n",
				SDLNet_GetError());
		//exit ( 2 ) ;
	}
	else
	{
		Printf ( "|| serving at port %d...\n", port ) ;
		SDLNet_TCP_AddSocket(socketset, self . socket);
	}
}

void Network::Disconnect ( peer *peer )
{
	if ( peer )
	{
		char key [ 256 ] ;
		char word [ 256 ] ;
		if ( peer -> timeout < SDL_GetTicks () )
			sprintf ( word, "%s", "lost" ) ;
		else
			sprintf ( word, "%s", "closed" ) ;

		Printf ( "|| %s connection to '%s'.\n", word, ResolveIP ( peer -> address ) ) ;
		
		sprintf ( key, "%s", peer -> key ) ;
		
		int n = FindPeerBySocketI ( peer -> socket ) ;

		SDLNet_TCP_DelSocket(socketset, peer -> socket);
		SDLNet_TCP_Close(peer -> socket);
		
		// remove buddies that can't be reached anymore
		for ( unsigned int i = 0 ; i < peer -> known_hosts . Count () ; i++ )
		{
			bool is_available = false ;
			host *host = peer -> known_hosts . ItemPointer ( i ) ;
			
			for ( unsigned int j = 0 ; j < peer_list -> Count () ; j++ )
			{
				struct peer *t_peer = peer_list -> ItemPointer ( j ) ;
				if ( strcmp ( t_peer -> key, peer -> key ) != 0 &&
						KeyIsKnownToPeer ( t_peer, host -> key ) )
					is_available = true ;
			}
			
			if ( ! is_available )
				DelBuddy ( host -> key ) ;
		}
		
		peer -> known_hosts . Clear () ;
		
		peer_list -> Remove ( n ) ;
		
		// notify all buddies that this peer has vanished
		for ( unsigned int i = 0 ; i < buddies -> Count () ; i++ )
		{
			SendMessage ( buddies -> Item ( i ) . address,
							key,
							PACKET_TYPE_DISCONNECTED ) ;
		}
		DelBuddy ( key ) ;
	}
}

/* obsolete */
void Network::StartAsServer ( int port ) {}
void Network::StartAsClient ( char *server, int port ) {}
int  Network::Connect () { return -1 ; }
void Network::Disconnect () {}
/* */

void Network::SendHandshake ( TCPsocket socket )
{
	char ip_string [ 256 ] ;
	struct peer *peer = FindPeerBySocket ( socket ) ;
	GetIPAsString ( &peer -> ip, (char*) ip_string ) ;
	SendMessage ( peer -> address, ip_string, PACKET_TYPE_HANDSHAKE ) ;
}

void Network::SendPing ( char *address )
{
	char time [ 256 ], dest [ 256 ] ;
	sprintf ( time, "%d", SDL_GetTicks () ) ;
	sprintf ( dest, "%s", address ) ;
	
	IPaddress ip ;
	Cached_SDLNet_ResolveHost ( &ip, address, 0 ) ;
	GetIPAsString ( &ip, dest ) ;
	
	int num_chars = SendMessage ( dest, time, PACKET_TYPE_PING ) ;
	printd ( "ping host %s (sent %d chars)\n", dest, num_chars ) ;
}

void Network::TCPAcceptEvent ( TCPsocket socket )
{
	IPaddress ip = *SDLNet_TCP_GetPeerAddress ( socket ) ;
	char name [ 256 ] ;

	ResolveIP ( &ip, (char*) name ) ;
	
	Printf ( "|| incoming connection from '%s' ( %08x ) on port %d...\n",
				name, ip . host, ip . port ) ;

	struct peer t_peer ;
	GetIPAsString ( &ip, (char*) t_peer . address ) ;
	t_peer . socket = socket ;
	t_peer . ip = ip ;
	t_peer . established = false ;
	t_peer . timeout = SDL_GetTicks () + CONNECTION_TIMEOUT * 1000 ;
	t_peer . born = SDL_GetTicks () ;
	t_peer . known_hosts . Clear () ;
	t_peer . outgoing_name [ 0 ] = '\0' ;
	t_peer . key [ 0 ] = '\0' ;
	t_peer . alias [ 0 ] = '\0' ;
	peer_list -> PushBack ( t_peer ) ;
	
	SDLNet_TCP_AddSocket(socketset, socket);
	
	Printf ( "|| sending %s a welcome message. waiting for respone...\n",
				name ) ;
	
	SendHandshake ( socket ) ;
}

bool Network::TCPReceiveEvent ( peer *peer )
{
	char string [ 8192 ] ;
	sprintf ( string, "%s", "" ) ;
	int string_len = SDLNet_TCP_Recv ( peer -> socket, string, 8192 ) ;
	//printd ( "got %d chars from peer '%s'.\n", string_len, peer->address ) ;
	
	if ( string_len > 0 && strlen ( string ) < 16 )
		return true ;
	
	if ( string_len > 0 )
	{
		struct packet packet ;
		
		char t_string [ 8192 ] ;
		sprintf ( t_string, "%s", string ) ;
		//GlDynArray < std::string > *args = GetArgList ( t_string, PACKET_DATA_DELIMITER ) ;
		GlDynArray < std::string > *args = new GlDynArray < std::string > ;
		
		int start = 0, stop = 0 ;
		for ( unsigned int i = 0 ; i < strlen ( t_string ) + 1 ; i++ )
		{
			if ( i == strlen ( t_string ) || t_string [ i ] == PACKET_DATA_DELIMITER )
			{
				stop = i ;
				char part [ strlen ( t_string ) ] ;
				sprintf ( part, "%s", &t_string [ start ] ) ;
				part [ stop - start ] = '\0' ;
				std::string submit = part ;
				args -> PushBack ( submit ) ;
				start = stop + 1 ;
			}
		}

		// invalid packet?
		if ( ! args || args -> Count () < 10 )
		{
			printd ( "received invalid packet (%d chars) from %s:\n'%s'\n",
						string_len, ResolveIP ( peer -> address ), string ) ;
			args -> Clear () ;
			delete args ;
			return true ;
		}
		
		// extract data
		packet . type      = atoi ( args -> Item ( 0 ) . c_str () ) ;
		sprintf ( packet . from, "%s", args -> Item ( 1 ) . c_str () ) ;
		sprintf ( packet . dest, "%s", args -> Item ( 2 ) . c_str () ) ;
		sprintf ( packet . from_key, "%s", args -> Item ( 3 ) . c_str () ) ;
		sprintf ( packet . dest_key, "%s", args -> Item ( 4 ) . c_str () ) ;
		sprintf ( packet . alias, "%s", args -> Item ( 5 ) . c_str () ) ;
		sprintf ( packet . hash, "%s", args -> Item ( 6 ) . c_str () ) ;
		packet . hops      = atoi ( args -> Item ( 7 ) . c_str () ) ;
		packet . max_hops  = atoi ( args -> Item ( 8 ) . c_str () ) ;
		sprintf ( packet . data, "%s", args -> Item ( 9 ) . c_str () ) ;
		
		args -> Clear () ;
		delete args ;
		args = NULL ;
		
		packet . hops++ ;
		
		if ( strcmp ( packet . from_key, peer -> key ) != 0 )
			printd ( "got packet from %s :\n%s\n", ResolveIP ( peer -> address ), string ) ;
		
		if ( packet . hops > packet . max_hops )
		{
			printd ( "received packet from %s is too old. hops = %d ; max = %d\n",
						ResolveIP ( packet . from ), packet . hops, packet . max_hops ) ;
			return true ;
		}
		
		if ( strcmp ( packet . from_key, self . key ) == 0 )
		{
			printd ( "received my own packet from peer %s. thrown away.\n'%s'\n",
						ResolveIP ( peer -> address ), string ) ;
			
			if ( strcmp ( packet . from_key, self . key ) == 0 &&
					( strcmp ( peer -> key, self . key ) == 0 ||
						strlen ( peer -> key ) == 0 ) )
			{
				printd ( "detected loop with myself. Disconnecting...\n" ) ;
				Disconnect ( peer ) ;
				return false ;
			}
			
			return true ;
		}
		
		// check known packet list for outdated ones...
		for ( unsigned int i = 0 ; i < known_packets -> Count () ; i++ )
		{
			known_packet *kp = known_packets -> ItemPointer ( i ) ;
			if ( kp -> timeout < SDL_GetTicks () )
			{
				known_packets -> Remove ( i ) ;
				i = 0 ; // reset counter and start over.. :)
			}
		}
		
		// check if we've already seen this packet!! IMPORTANT!!
		for ( unsigned int i = 0 ; i < known_packets -> Count () ; i++ )
		{
			known_packet *kp = known_packets -> ItemPointer ( i ) ;
			
			if ( strcmp ( kp -> hash, packet . hash ) == 0 )
			{
				printd ( "packet from %s already seen. thrown away...\n'%s'\n",
							ResolveIP ( packet . from ), string ) ;
				return true ;
			}
		}
		
		// add this packet info to our list
		if ( strlen ( packet . hash ) )
		{
			known_packet kp ;
			sprintf ( kp . hash, "%s", packet . hash ) ;
			kp . timeout = SDL_GetTicks () + HOST_CACHE_TIMEOUT * 1000 ;
			known_packets -> PushBack ( kp ) ;
		}
		
		// redirect if needed
		if ( strlen ( self . key ) && strlen ( packet . dest_key ) &&
				strcmp ( packet . dest_key, self . key ) != 0 )
		{
			packet . hops++ ;
			for ( unsigned int i = 0 ; i < peer_list -> Count () ; i++ )
			{
				struct peer *t_peer = peer_list -> ItemPointer ( i ) ;
				
				if ( strcmp ( t_peer -> key, peer -> key ) != 0 &&
						( KeyIsKnownToPeer ( t_peer, packet . dest_key ) ||
					          strcmp ( t_peer -> key, packet . dest_key ) == 0 ) )
				{
					char from [ 256 ], to [ 256 ], dest [ 256 ] ;
					sprintf ( from, "%s", ResolveIP ( packet . from ) ) ;
					sprintf ( to, "%s", ResolveIP ( t_peer -> address ) ) ;
					sprintf ( dest, "%s", ResolveIP ( packet . dest ) ) ;
					printd ( "redirecting packet from %s to %s (destined for %s). data = '%s'\n",
							from,
							to,
							dest,
							packet . data ) ;
					SendPacket ( t_peer, &packet ) ;
				}
			}
			printd ( "packet not redirected.\n" ) ;
			//return true ;
		}
		
		if ( packet . type == PACKET_TYPE_HANDSHAKE )
		{	// this is a greeting! setup needed data...
			printd ( "i'm seen as %s\n", ResolveIP ( packet . data ) ) ;
			sprintf ( peer -> outgoing_name, "%s", packet . data ) ;
			
			if ( strlen ( packet . from ) < 1 )
			{
				// let's tell our peer how we see it!
				SendHandshake ( peer -> socket ) ;
				// and write it down for ourself...
				GetIPAsString ( &peer -> ip, (char*) packet . from ) ;
			}
			
			if ( strcmp ( peer -> address, packet . from ) == 0 )
			{
				sprintf ( peer -> alias, "%s", packet . alias ) ;
				sprintf ( peer -> key, "%s", packet . from_key ) ;
				
				printd ( "set alias '%s' and key for peer %s.\n",
							peer -> alias, ResolveIP ( peer -> address ) ) ;
			}
			
			// add host to buddy list
			AddBuddy ( packet . alias, packet . from_key, packet . from ) ;
			
			return true ;
		}
		else if ( strcmp ( peer -> key, packet . from_key ) != 0 &&
					! KeyIsKnownToPeer ( peer , packet . from_key ) )
		{	// add unknown host to our list
			host host ;
			sprintf ( host . alias, "%s", packet . alias ) ;
			sprintf ( host . key, "%s", packet . from_key ) ;
			sprintf ( host . address, "%s", packet . from ) ;
			host . timeout = SDL_GetTicks () + CONNECTION_TIMEOUT * 1000 ;
			
			peer -> known_hosts . PushBack ( host ) ;
			
			printd ( "added host %s to known-hosts-list for peer %s.\n",
						ResolveIP ( host . address ), ResolveIP ( peer -> address ) ) ;
			// add host to buddy list
			AddBuddy ( packet . alias, packet . from_key, packet . from ) ;
		}
		
		// check for changes in host-data...
		if ( strcmp ( peer -> key, packet . from_key ) == 0 )
		{	// we know this peer. check for changed data
			if ( strcmp ( peer -> alias, packet . alias ) != 0 )
			{	// peer changed alias name!
				Printf ( "|| peer %s (%s) is now known as %s.\n",
							peer -> alias, ResolveIP ( peer -> address ),
							packet . alias ) ;
				// change name in buddy list
				ChangeBuddyName ( packet . from_key, packet . alias ) ;
				
				// write it down
				sprintf ( peer -> alias, "%s", packet . alias ) ;
			}
		}
		else if ( KeyIsKnownToPeer ( peer, packet . from_key ) )
		{	// we know this host. check for changed data
			GlDynArray < host > *known_hosts = &peer -> known_hosts ;
			
			for ( unsigned int i = 0 ; i < known_hosts -> Count () ; i++ )
			{	// get our foreign host
				host *host = known_hosts -> ItemPointer ( i ) ;
				if ( strcmp ( host -> key, packet . from_key ) == 0 &&
						strcmp ( host -> alias, packet . alias ) != 0 )
				{	// host changed alias name!
					Printf ( "|| %s (%s) is now known as %s.\n",
								host -> alias, ResolveIP ( packet . from ),
								packet . alias ) ;
					// change name in buddy list
					ChangeBuddyName ( host -> key, packet . alias ) ;
					
					// write it down
					sprintf ( host -> alias, "%s", packet . alias ) ;
				}
			}
		}
		
		// check if it's a ping :)
		if ( packet . type == PACKET_TYPE_PING )
		{	// send alive signal if it's for me
			if ( strcmp ( packet . dest_key, self . key ) == 0 )
			{
				sprintf ( packet . dest, "%s", packet . from ) ;
				sprintf ( packet . from, "%s", peer -> outgoing_name ) ;
				sprintf ( packet . dest_key, "%s", packet . from_key ) ;
				sprintf ( packet . from_key, "%s", self . key ) ;
				sprintf ( packet . alias, "%s", self . alias ) ;
				sprintf ( packet . hash, "%s", CreateKey () . c_str () ) ;
				packet . type = PACKET_TYPE_ALIVE ;
				printd ( "return ping to %s. current ticks = %d ; data = %s\n",
							packet . dest, SDL_GetTicks (), packet . data ) ;
				SendPacket ( peer, &packet ) ;
				return true ;
			}
		}
		
		// check if it's a alive signal - probably to update a buddy etc..
		if ( /*packet . type == PACKET_TYPE_ALIVE*/ true )
		{
			host *buddy = NULL ;
			
			long int act_time = SDL_GetTicks () ;
			
			buddy = FindBuddyByKey ( packet . from_key ) ;
			
			if ( packet . type == PACKET_TYPE_ALIVE &&
					( strcmp ( packet . dest_key, self . key ) == 0 ) && buddy )
			{
				long int time = act_time - atol ( packet . data ) ;
				printd ( "host %s: hops = %d ; response time = %ld\n",
							ResolveIP ( packet . from ),
							packet . hops,
							time ) ;
				if ( time >= 0 )
					buddy -> speed = time ;
				else
					printd ( "sdl ticks = %ld ; data = %s\n", act_time, packet . data ) ;
			}
			
			if ( buddy )
			{
				int time_left = ( buddy -> timeout - act_time ) / 1000 ;
				if ( time_left < BUDDY_TIMEOUT - PING_INTERVAL / 2 )
					SendPing ( buddy -> address ) ;
				
				printd ( "refresh buddy timout of %s.\n", packet . alias ) ;
				buddy = RefreshBuddy ( packet . from_key ) ;
			}
			else
			{
				printd ( "found unknown buddy %s.\n", packet . alias ) ;
				buddy = AddBuddy ( packet . alias, packet . from_key, packet . from ) ;
			}
			
			if ( strcmp ( packet . from_key, peer -> key ) == 0 )
			{
				printd ( "refresh peer timout of %s.\n", peer -> address ) ;
				peer -> timeout = act_time + CONNECTION_TIMEOUT * 1000 ;
			}
			
			if ( packet . type == PACKET_TYPE_ALIVE )
			{
				for ( unsigned int p = 0 ; p < peer_list -> Count () ; p++ )
				{
					struct peer *other_peer = peer_list -> ItemPointer ( p ) ;
					if ( strcmp ( other_peer -> key, peer -> key ) != 0 )
					{
						if ( packet . hops < packet . max_hops )
						{
							printd ( "send copy of packet from %s to %s.\n",
										ResolveIP ( packet . from ),
										ResolveIP ( other_peer -> address ) ) ;
							SendPacket ( other_peer, &packet ) ;
						}
						else
						{
							printd ( "won't send copy of packet from %s to %s. hops %d >= max_hops %d\n",
										ResolveIP ( packet . from ),
										ResolveIP ( other_peer -> address ),
										packet . hops, packet . max_hops ) ;
						}
					}
				}
				
				//return true ;
			}
		}
		
		// check if someone disconnected
		if ( packet . type == PACKET_TYPE_DISCONNECTED )
		{
			char alias [ 256 ] ;
			char address [ 256 ] ;
			
			bool found = false ;
			for ( unsigned int i = 0 ; i < buddies -> Count () ; i++ )
			{
				host *buddy = buddies -> ItemPointer ( i ) ;
				if ( strcmp ( buddy -> key, packet . data ) == 0 )
				{
					found = true ;
					sprintf ( alias, "%s", buddy -> alias ) ;
					sprintf ( address, "%s", buddy -> address ) ;
				}
			}
			
			if ( found )
			{
				printd ( "%s (%s) disconnected from us.\n",
							alias, ResolveIP ( address ) ) ;
				DelBuddy ( packet . data ) ;
			}
			else if ( strcmp ( packet . data, self . key ) == 0 )
			{
				printd ( "sending alive signal to everyone. i'm not gone yet :)\n" ) ;
				for ( unsigned int i = 0 ; i < buddies -> Count () ; i++ )
					SendMessage ( buddies -> Item ( i ) . address,
									"ALIVE", PACKET_TYPE_ALIVE ) ;
				return true ;
			}
		}
		
		// TODO
		
		if ( packet . type == PACKET_TYPE_MESSAGE )
			Printf ( "%s (%s): %s\n", packet . alias,
						ResolveIP ( packet . from ), packet . data ) ;
		else if ( packet . type == PACKET_TYPE_PMESSAGE &&
					strcmp ( packet . dest_key, self . key ) == 0 )
		{
			Printf ( "PM from %s (%s): %s\n", packet . alias,
			ResolveIP ( packet . from ), packet . data ) ;
			return true ;
		}
		
		for ( unsigned int i = 0 ; i < peer_list -> Count () ; i++ )
		{
			struct peer *t_peer = peer_list -> ItemPointer ( i ) ;
			
			if ( strcmp ( t_peer -> key, peer -> key ) != 0 )
			{
				if ( packet . type == PACKET_TYPE_MESSAGE )
				{
					sprintf ( packet . dest, "%s", t_peer -> address ) ;
					sprintf ( packet . dest_key, "%s", t_peer -> key ) ;
				}
				if ( packet . hops < packet . max_hops )
					SendPacket ( t_peer, &packet ) ;
				else
					printd ( "won't resend packet from %s. hops %d >= max_hops %d\n",
								ResolveIP ( packet . from ), packet . hops, packet . max_hops ) ;
			}
		}

		return true ;
	}
	else
	{
		Printf ( "|| connection to '%s' has died.\n", ResolveIP ( peer -> address ) ) ;
		TCPCloseEvent ( peer ) ;
		
		return false ;
	}
}

void Network::TCPCloseEvent ( peer *peer )
{
	Disconnect ( peer ) ;
}


// buddy list thingies
Network::host *Network::FindBuddyByKey ( char *buddy_key )
{
	for ( unsigned int i = 0 ; i < buddies -> Count () ; i++ )
	{
		if ( strcmp ( buddies -> ItemPointer ( i ) -> key, buddy_key ) == 0 )
		{
			//printd ( "buddy %s already in list.\n",
			//			buddies -> ItemPointer ( i ) -> alias ) ;
			return buddies -> ItemPointer ( i ) ;
		}
	}
	
	return NULL ;
}
	
Network::host *Network::AddBuddy ( char *alias, char *key, char *from_address )
{
	if ( strcmp ( key, self . key ) == 0 )
	{
		printd ( "won't add myself to my buddy list!\n" ) ;
		return NULL ;
	}
	
	if ( FindBuddyByKey ( key ) )
	{
		printd ( "buddy %s already in list.\n", alias ) ;
		return NULL ;
	}
	
	host host ;
	sprintf ( host . alias, "%s", alias ) ;
	sprintf ( host . key, "%s", key ) ;
	sprintf ( host . address, "%s", from_address ) ;
	host . timeout = SDL_GetTicks () + BUDDY_TIMEOUT * 1000 ; // time to respond
	host . speed = 0 ; // respond time in ms
	
	buddies -> PushBack ( host ) ;
	
	Printf ( "|| %s (%s) has entered our world.\n", 
				host . alias, ResolveIP ( host . address ) ) ;
	
	SendPing ( host . address ) ;
	
	return RefreshBuddy ( key ) ;
}

bool Network::DelBuddy ( char *key )
{
	for ( unsigned int i = 0 ; i < buddies -> Count () ; i++ )
	{
		if ( strcmp ( buddies -> Item ( i ) . key, key ) == 0 )
		{
			Printf ( "|| %s (%s) left us.\n", 
						buddies -> Item ( i ) . alias,
						ResolveIP ( buddies -> Item ( i ) . address ) ) ;
			buddies -> Remove ( i ) ;
			return true ;
		}
	}
	
	return false ;
}

Network::host *Network::RefreshBuddy ( char *key )
{
	printd ( "searching for buddy with key '%s'\n", key ) ;
	for ( unsigned int i = 0 ; i < buddies -> Count () ; i++ )
	{
		if ( strcmp ( buddies -> Item ( i ) . key, key ) == 0 )
		{
			buddies -> ItemPointer ( i ) -> timeout = SDL_GetTicks () + BUDDY_TIMEOUT * 1000 ;
			printd ( "found %s (%s). updated timeout.\n",
						buddies -> ItemPointer ( i ) -> alias,
						ResolveIP ( buddies -> ItemPointer ( i ) -> address ) ) ;
			return buddies -> ItemPointer ( i ) ;
		}
	}
	printd ( "no buddy with key '%s' found!\n", key ) ;
	return NULL ;
}

Network::host *Network::ChangeBuddyName ( char *key, char *new_name )
{
	for ( unsigned int i = 0 ; i < buddies -> Count () ; i++ )
	{
		if ( strcmp ( buddies -> Item ( i ) . key, key ) == 0 )
		{
			Printf ( "|| %s is now called %s.\n",
						buddies -> Item ( i ) . alias, new_name ) ;
			sprintf ( buddies -> ItemPointer ( i ) -> alias, "%s", new_name ) ;
			UpdateBuddyList () ;
			return buddies -> ItemPointer ( i ) ;
		}
	}
	
	return NULL ;
}

void Network::UpdateBuddyList ()
{	// update buddy list!

#ifdef __NCURSES_SUPPORT__
	if ( ! nconsole && ! buddy_list ) return ;
#else
	if ( ! buddy_list ) return ;
#endif
	
	GlDynArray < std::string > t_buddies ;
	t_buddies . Clear () ;
	for ( unsigned int i = 0 ; i < buddies -> Count () ; i++ )
	{
		struct host *user = buddies -> ItemPointer ( i ) ;
		
		char line [ 256 ] ;
		int time_left = ( user -> timeout - SDL_GetTicks () ) / 1000 ;
		sprintf ( line, "%s", user -> alias ) ;
		
		//if ( time_left <= 10 || user -> speed == 0 )
		//	sprintf ( line, "%s (-%ds)", line, time_left ) ;
		//else
		//	sprintf ( line, "%s (%dms)", line, user -> speed ) ;
		
		char middle [ 5 ], line2 [ 256 ] ;
		sprintf ( middle, "%s", "" ) ;
		sprintf ( line2, "%s", "" ) ;
		if ( time_left <= max ( BUDDY_TIMEOUT / 3, 10 ) )
		{
			sprintf ( line2, "-%ds", time_left ) ;
			sprintf ( middle, "%s", "/" ) ;
		}
		if ( user -> speed > 0 && user -> speed < (uint) BUDDY_TIMEOUT * 1024 )
			sprintf ( line2, "%s%s~%dms", line2, middle, user -> speed ) ;
		if ( strlen ( line2 ) )
			sprintf ( line, "%s (%s)", line, line2 ) ;
		
		t_buddies . PushBack ( line ) ;
	}
	
#ifdef __NCURSES_SUPPORT__
	if ( nconsole )
		nconsole -> UpdateBuddyList ( &t_buddies ) ;
#endif
	
	if ( buddy_list )
	{
		unsigned int users_in_buddy_list = t_buddies . Count () ;
		unsigned int max_lines = buddy_list -> NumLines () ;
		unsigned int uid = 0 ;
		for ( unsigned int i = 0 ; i < max_lines ; i++ )
		{
			if ( uid < users_in_buddy_list )
			{
				buddy_list -> SetLine ( i, t_buddies [ uid ] ) ;
				
				uid++ ;
			}
			else
				buddy_list -> SetLine ( i, "" ) ;
		}
	}
}
// end of buddy list thingies


void Network::ProcessEvents ()
{
	static unsigned int users_in_buddy_list = buddies -> Count () ;
	
	// wait a moment for events
	SDLNet_CheckSockets(socketset, 0);
	
	if ( self . socket )
	{
		// Check for new connections
		while ( SDLNet_SocketReady(self . socket) )
		{
			TCPsocket socket = SDLNet_TCP_Accept ( self . socket ) ;
			if ( socket != NULL )
			{
				TCPAcceptEvent ( socket ) ;
			}
		}
	}
	
	bool send_ping = false, update = false ;
	static Uint32 last_ping, this_ping = last_ping = SDL_GetTicks () ;
	static int secs_counted = 0 ;
	this_ping = SDL_GetTicks () ;
	if ( this_ping - last_ping > 1000 )
	{
		update = true ;
		secs_counted++ ;
		last_ping = this_ping ;
	}
	
	if ( secs_counted > PING_INTERVAL )
	{
		send_ping = true ;
		secs_counted = 0 ;

		for ( unsigned int i = 0 ; i < dnscache . Count () ; i++ )
		{
			if ( dnscache [ i ] . timeout < SDL_GetTicks () )
			{
				//printd ( "remove entry %d from dns-cache.\n", i ) ;
				dnscache . Remove ( i ) ;
				i = 0 ;
			}
		}
	}
	
	for ( unsigned int i = 0 ; i < peer_list -> Count () ; i++ )
	{
		bool still_active = true ;
		peer *peer = peer_list -> ItemPointer ( i ) ;
		
		// Check for events on existing client
		if ( still_active && SDLNet_SocketReady(peer -> socket) )
			still_active = TCPReceiveEvent ( peer ) ;
		
		if ( still_active )
		{
			int time_left = ( peer -> timeout - SDL_GetTicks () ) / 1000 ;
			if ( send_ping && time_left < 10 )
				SendPing ( peer -> address ) ;
			
			// has the connection timed out?
			if ( peer -> timeout < SDL_GetTicks () )
			{
				Disconnect ( peer ) ;
				continue ;
			}
		}
		else // connection died
			break ;
	}
	
	// remove stale buddies
	for ( unsigned int i = 0 ; i < buddies -> Count () ; i++ )
	{
		if ( buddies -> ItemPointer ( i ) -> timeout < SDL_GetTicks () )
		{
			DelBuddy ( buddies -> ItemPointer ( i ) -> key ) ;
			i = 0 ; // recount
		}
		else if ( send_ping )
			SendPing ( buddies -> ItemPointer ( i ) -> address ) ;
	}
	
	if ( users_in_buddy_list != buddies -> Count () || update )
	{	// update buddy list!
		users_in_buddy_list = buddies -> Count () ;
		UpdateBuddyList () ;
	}
	
#ifdef __NCURSES_SUPPORT__
	static Uint32 this_update, last_update = SDL_GetTicks () ;
	this_update = SDL_GetTicks () ;
	if ( nconsole && this_update - last_update > 100 )
	{
		nconsole -> Update () ;
		last_update = this_update ;
	}
#endif
}

char *Network::StripHeadAndTail ( char *text )
{
	int  length = strlen ( text ) ;
	int  start, end, pos ;
	
	for ( pos = 0 ; pos < length ; pos++ )
		if ( text [ pos ] == 9 )
			text [ pos ] = ' ' ;
	
	for ( pos = 0 ; pos < length ; pos++ )
		if ( text [ pos ] != ' ' )
			break ;
	
	start = pos ;
	
	for ( pos = length - 1 ; pos > 0 ; pos-- )
	{
		if ( text [ pos ] != ' ' )
			break ;
	}
	
	end = pos + 1 ;
	
	for ( pos = start ; pos < end ; pos++ )
		text [ pos - start ] = text [ pos ] ;
	for ( pos = pos - start ; pos < length ; pos++ )
		text [ pos ] = '\0' ;
	
	return text ;
}

char *Network::StripMarks ( char *text )
{
	int old_pos = 0 ;
	
	for ( unsigned int pos = 0 ; pos < strlen ( text ) ; pos ++ )
	{
		if ( !( pos == 0 && text [ pos ] == '"' ) &&
				!( pos == strlen ( text ) - 1 && text [ pos ] == '"' ) )
			text [ old_pos++ ] = text [ pos ] ;
	}
	
	text [ old_pos ] = '\0' ;
	
	return text ;
}

GlDynArray < std::string > *Network::GetArgList ( char *string,
													char delimiter )
{
	GlDynArray < std::string > *bucket = new GlDynArray < std::string > () ;
	
	bucket -> Clear () ;
	unsigned int pos = 0 ;
	
	bool marks_found = FALSE ;
	char eop = 255 ; // end of part mark
	
	//printd ( "args =" ) ;
	
	char *copy = (char*) malloc ( sizeof (char) * 8192 ) ;
	sprintf ( copy, "%s", string ) ;
	
	while ( pos < strlen ( copy ) )
	{
		std::string arg ;
		for ( ; pos < strlen ( copy ) ; pos++ )
		{
			if ( copy [ pos ] == '"' ) marks_found = ! marks_found ;
			if ( copy [ pos ] == delimiter && ! marks_found ) break ;
			arg += copy [ pos ] ;
		}
		arg += '\0' ;
		if ( pos < strlen ( copy ) )
			copy [ pos ] = eop ;
		
		if ( (arg . c_str ()) [ 0 ] == eop ) arg = arg . c_str () + 1 ;

		arg = StripMarks ( StripHeadAndTail ( (char*) arg . c_str () ) ) ;
		
		bucket -> PushBack ( arg ) ;
		
		//printd ( " '%s'", (char*) bucket [ bucket . Count () - 1 ] . c_str () ) ;
	}
	
	//printd ( "\n" ) ;
	
	free ( copy ) ;
	
	return bucket ;
}

// returns true if it could be interpreted, otherwise it's a global message
bool Network::ProcessMessage ( peer *peer, char *t_message )
{
	char message [ 8192 ] ;
	sprintf ( message, "%s", t_message ) ;
	
	GlDynArray < std::string > *arg_list = GetArgList ( message ) ;
	if ( arg_list -> Count () > 0 )
	{
		// display about header: //about
		if ( strcmp ( arg_list -> Item ( 0 ) . c_str (), "//about" ) == 0 )
		{
			Printf ( "|| welcome to the gemhun network console !\n" ) ;
			Printf ( "||\n" ) ;
			
			return true ;
		}
		
		// display command-line help: //help
		if ( strcmp ( arg_list -> Item ( 0 ) . c_str (), "//help" ) == 0 )
		{
			if ( arg_list -> Count () == 1 )
			{	// display global help
				Printf ( "\n|| global help follows:\n" ) ;
				Printf ( "||\n" ) ;
				Printf ( "|| connect hostname [port]  -  connect with specified server at port ( defaults to 5555 )\n" ) ;
				Printf ( "|| disconnect hostname      -  disconnect from specified server\n" ) ;
				Printf ( "|| alias new_username       -  change username to new_username\n" ) ;
				Printf ( "|| pm username message      -  send private message to user\n" ) ;
				Printf ( "|| sockets                  -  list all connections\n" ) ;
				Printf ( "||\n" ) ;
				Printf ( "|| help                     -  this help :)\n" ) ;
				Printf ( "|| help command             -  help on specified command\n" ) ;
				Printf ( "||\n" ) ;
				Printf ( "|| when using a command on the console do it like this:\n" ) ;
				Printf ( "|| //alias superman\n" ) ;
				Printf ( "\n" ) ;
			}
			else if ( arg_list -> Count () == 2 )
			{	// display command-specific help
				char *command = (char*) arg_list -> Item ( 1 ) . c_str () ;
				Printf ( "\n|| help for '%s' command:\n", command ) ;
				Printf ( "\n" ) ;
			}
			else
			{
				Printf ( "\n|| help command can be used in two ways:\n" ) ;
				Printf ( "|| - give a name for the command you want help for\n" ) ;
				Printf ( "|| - don't specify anything and you'll get the full list\n\n" ) ;
			}
			
			return true ;
		}
		
		// connect to peer: //connect <hostname> [<port>]
		if ( strcmp ( arg_list -> Item ( 0 ) . c_str (), "//connect" ) == 0 )
		{
			if ( arg_list -> Count () >= 2 && arg_list -> Count () <= 3 )
			{
				char host [ 256 ] ;
				sprintf ( host, "%s", (char*) arg_list -> Item ( 1 ) . c_str () ) ;
				int port = 5555 ;
				
				if ( arg_list -> Count () == 3 )
					port = atoi ( arg_list -> Item ( 2 ) . c_str () ) ;
				
				ConnectTo ( host, port ) ;
			}
			else
			{
				Printf ( "\n|| connect command needs ca. 2 parameters:\n" ) ;
				Printf ( "|| - the name or the ip-address of the computer\n" ) ;
				Printf ( "|| - the port on which your partner listens on (optional / defaults to 5555)\n" ) ;
				Printf ( "|| example:\n" ) ;
				Printf ( "||   connect mycomputer\n" ) ;
				Printf ( "||   or\n" ) ;
				Printf ( "||   connect mycomputer 5555\n\n" ) ;
			}
			
			return true ;
		}
		
		// disconnect from peer: //disconnect <hostname>
		if ( strcmp ( arg_list -> Item ( 0 ) . c_str (), "//disconnect" ) == 0 )
		{
			if ( arg_list -> Count () >= 2 && arg_list -> Count () <= 3 )
			{
				char host [ 256 ] ;
				sprintf ( host, "%s", (char*) arg_list -> Item ( 1 ) . c_str () ) ;
				
				IPaddress ip ;
				Cached_SDLNet_ResolveHost ( &ip, host, 0 ) ;
				GetIPAsString ( &ip, host ) ;
				
				struct peer *peer = NULL ;

				for ( unsigned int i = 0 ; i < peer_list -> Count () && peer == NULL ; i++ )
				{
					char shortname [ 256 ], longname [ 256 ] ;
					sprintf ( longname, "%s", ResolveIP ( peer_list -> ItemPointer ( i ) -> address ) ) ;
					sprintf ( shortname, "%s", longname ) ;
					if ( index ( shortname, '.' ) )
						*index ( shortname, '.' ) = '\0' ;
					
					if ( strcmp ( peer_list -> ItemPointer ( i ) -> alias, host ) == 0 ||
						strcmp ( peer_list -> ItemPointer ( i ) -> address, host ) == 0 ||
						strcmp ( longname, host ) == 0 || strcmp ( shortname, host ) == 0 )
						peer = peer_list -> ItemPointer ( i ) ;
				}

				if ( peer )
					Disconnect ( peer ) ;
			}
			else
			{
				Printf ( "\n|| disconnect command needs a hostname as parameter:\n" ) ;
				Printf ( "|| - the name or the ip-address of the computer\n" ) ;
				Printf ( "|| example:\n" ) ;
				Printf ( "||   disconnect mycomputer\n\n" ) ;
			}
			
			return true ;
		}
		
		// change username: //alias <new username>
		if ( strcmp ( arg_list -> Item ( 0 ) . c_str (), "//alias" ) == 0 )
		{
			if ( arg_list -> Count () == 2 )
			{
				sprintf ( self . alias, "%s", arg_list -> Item ( 1 ) . c_str () ) ;
				SendMessage ( NULL, "ALIVE", PACKET_TYPE_ALIVE ) ;
			}
			else
			{
				Printf ( "\n|| alias command needs exactly 1 parameter:\n" ) ;
				Printf ( "|| - the new name (alias) of the computer\n" ) ;
				Printf ( "|| example:\n  alias johnboy\n" ) ;
				Printf ( "|| this would display yourself as johnboy.\n" ) ;
				Printf ( "|| note: you could also supply this: \"i'm the master\"\n\n" ) ;
			}
			
			return true ;
		}
		
		// list sockets: //sockets
		if ( strcmp ( arg_list -> Item ( 0 ) . c_str (), "//sockets" ) == 0 )
		{
			if ( arg_list -> Count () == 1 )
			{
				if ( peer_list -> Count () == 0 )
					Printf ( "\n|| there are no open connections.\n" ) ;
				else
					Printf ( "\n|| found %d open connections. listing...\n",
								peer_list -> Count () ) ;
				
				for ( unsigned int i = 0 ; i < peer_list -> Count () ; i++ )
				{
					struct peer *t_peer = peer_list -> ItemPointer ( i ) ;
					int s = FindPeerBySocketI ( t_peer -> socket ) ;
					
					int speed = 0, secs = 0 ;
					host *buddy = NULL ;
					for ( unsigned int i = 0 ; i < buddies -> Count () && buddy == NULL ; i++ )
					{
						if ( strcmp ( buddies -> ItemPointer ( i ) -> key, t_peer -> key ) == 0 )
							buddy = buddies -> ItemPointer ( i ) ;
					}
					
					if ( buddy )
					{
						speed = buddy -> speed ;
						secs = ( t_peer -> timeout - SDL_GetTicks () ) / 1000 ;
					}
					
					Printf ( "|| socket %d: %s:%d %dms -%dsecs\n",
								s,
								ResolveIP ( t_peer -> address ),
								t_peer -> ip . port,
								speed, secs ) ;
				}
				Printf ( "\n" ) ;
			}
			else
			{
				Printf ( "\n|| sockets command doesn't need any arguments.\n\n" ) ;
			}
			
			return true ;
		}
		
		// send private message: //pm <username> message
		if ( strcmp ( arg_list -> Item ( 0 ) . c_str (), "//pm" ) == 0 )
		{
			if ( arg_list -> Count () >= 3 )
			{
				char name [ 256 ] ;
				sprintf ( name, "%s", arg_list -> Item ( 1 ) . c_str () ) ;
				
				char pm [ 256 ] ;
				char *string = strstr ( message, name ) + strlen ( name ) ;
				while ( string [ 0 ] == ' ' )
					string++ ;
				strcpy ( pm, string ) ;
				
				char address [ 256 ] ;
				address [ 0 ] = '\0' ;
				for ( unsigned int i = 0 ; i < buddies -> Count () ; i++ )
				{
					if ( strcmp ( buddies -> Item ( i ) . alias, name ) == 0 )
						sprintf ( address, "%s", buddies -> Item ( i ) . address ) ;
				}
				
				if ( strlen ( address ) ) 
					SendMessage ( address, pm, PACKET_TYPE_PMESSAGE ) ;
				else
					Printf ( "|| sorry, can't send pm to unknown buddy %s!\n",
							name ) ;
				return true ;
			}
			else
			{
				Printf ( "\n|| pm command needs 2 parameters:\n" ) ;
				Printf ( "|| - the alias name of the destination computer\n" ) ;
				Printf ( "|| - the message to send to\n" ) ;
				Printf ( "|| example:\n  //pm johnboy greetingz sir :)\n" ) ;
				Printf ( "|| this would send johnboy a nice message.\n\n" ) ;
			}
			
			return true ;
		}

		// send ping to host: //ping <hostname or ip-address>
		if ( strcmp ( arg_list -> Item ( 0 ) . c_str (), "//ping" ) == 0 )
		{
			if ( arg_list -> Count () == 2 )
			{
				char address [ 256 ] ;
				sprintf ( address, "%s", arg_list -> Item ( 1 ) . c_str () ) ;
				
				if ( strlen ( address ) ) 
					SendPing ( address ) ;
				
				return true ;
			}
			else
			{
				Printf ( "\n|| ping command needs a parameter:\n" ) ;
				Printf ( "|| - the hostname or ip-address to ping to\n" ) ;
				Printf ( "|| example:\n  //ping mycomputer\n" ) ;
				Printf ( "|| this would ping mycomputer.\n\n" ) ;
			}
			
			return true ;
		}

		// i think nobody handled this command :(
		if ( strncmp ( arg_list -> Item ( 0 ) . c_str (), "//", 2 ) == 0 )
		{
			Printf ( "|| sorry, this failed. see //help for some hints...\n" ) ;

			return true ;
		}
	}

	// we don't know how to handle this - give it away :)
	SendMessage ( NULL, message, PACKET_TYPE_MESSAGE ) ;
	
	return false ;
}

void Network::UpdateNConsole ()
{
#ifdef __NCURSES_SUPPORT__
	if ( nconsole )
		nconsole -> Update () ;
#endif
}

void Network::Quit ()
{
	peer_list -> Clear () ; delete peer_list ;
	known_packets -> Clear () ; delete known_packets ;
	buddies -> Clear () ; delete buddies ;
	
	console = NULL ;
	buddy_list = NULL ;
	
#ifdef __NCURSES_SUPPORT__
	if ( nconsole )
	{
		delete nconsole ;
		nconsole = NULL ;
	}	
#endif
	
	if ( self . socket != NULL )
	{
		SDLNet_TCP_Close(self . socket);
		self . socket = NULL;
	}
	if ( socketset != NULL )
	{
		SDLNet_FreeSocketSet(socketset);
		socketset = NULL;
	}

	connectto_list . Clear () ;

	dnscache . Clear () ;

	SDLNet_Quit();
}

#endif
